/**
 * AutoRegister
 *
 * Auto register angular modules that are loaded so all you have to do is
 * import them as needed and don't need to manually add angular dependencies
 *
 * tl;dr you don't ever have to do angular.module 'app', [...stuff], instead
 * just
 *
 * 		import 'what-you-want';
 *
 * and any angular modules imported are automatically loaded for you as
 * a dependency of your application
 *
 * @todo move to node module / bower component
 */
export class AutoRegister {
  private modules: string[] = [];
  private appModuleName: string;

  constructor(appModuleName = 'app') {
    this.appModuleName = appModuleName;
    this.registerModule();
    this.stubModuleFunction();
  }

  /**
   * Register our main application module that everything gets bound to
   */
  private registerModule() {
    angular.module(this.appModuleName, this.modules);
  }

  /**
   * Stub angular.module to listen to registered modules
   * and automatically register them for the app module
   */
  private stubModuleFunction() {
    const originalAngularModuleFunction = angular.module;

    angular.module = (moduleName: string, dependencies: string[], ...otherArgs: any[]) => {
      // If the module being registered is not yet in our app module
      // dependencies, add it
      if (dependencies && this.modules.indexOf(moduleName) === -1) {
        this.modules.push(moduleName);
      }
      return originalAngularModuleFunction
        .call(angular, moduleName, dependencies, ...otherArgs);
    };
  }
}

export default new AutoRegister('app');
