// TODO: separate test dependencies
import '@host';

import 'angular-mocks';
import 'mocha';

import { expect } from 'chai';
import { ObjectUtils } from '../object';
import * as _ from 'lodash';

describe('objectUtils', () => {
  beforeEach(angular.mock.module('app'));

  describe('clear', () => {

    it('should return the original object/array reference', inject((objectUtils: ObjectUtils) => {
      const object = { foo: 'bar' };
      const returned = objectUtils.clear(object);

      expect(returned).to.equal(object);
    }));

    it('should clear all object keys', inject((objectUtils: ObjectUtils) => {
      const object = { foo: 'bar' };
      objectUtils.clear(object);

      // the object should have no keys
      expect(_.size(object)).to.equal(0);
    }));

    it('should clear all object keys', inject((objectUtils: ObjectUtils) => {
      const array = [1, 2, 3];
      objectUtils.clear(array);

      // the object should have no items
      expect(_.size(array)).to.equal(0);
    }));
  });

  describe('remove', () => {

    it('should return the original array reference', inject((objectUtils: ObjectUtils) => {
      const array: any[] = [1, 2, 3];
      const returned = objectUtils.clear(array);

      expect(returned).to.equal(array);
    }));

    it('should remove an item from an array', inject((objectUtils: ObjectUtils) => {
      const array = [1, 2, 3, 2];
      const returned = objectUtils.remove(array, 2);

      // the original array should not contain the removed value anymore
      expect(returned).to.not.contain(2);
    }));
  });

  describe('replace', () => {

    it('should return the original object reference', inject((objectUtils: ObjectUtils) => {
      const obj = {};
      const returned = objectUtils.replace(obj, { foo: 'bar' });
      expect(returned).to.equal(obj);
    }));

    it('should replace all keys and values of one object to another', inject((objectUtils: ObjectUtils) => {
      const obj1: any = { foo: 'bar' };
      const obj2 = { bar: 'foo' };

      objectUtils.replace(obj1, obj2);

      expect(obj1.foo).to.be.undefined;
      expect(obj1.bar).to.equal('foo');
    }));

    it('shoudl replace all array items of one array to another', inject((objectUtils: ObjectUtils) => {
      const arr1 = [1, 2, 3];
      const arr2 = [0, 1, 2];

      const returned = objectUtils.replace(arr1, arr2);

      expect(returned).to.equal(arr1);
      for (const item of arr2) {
        expect(arr1[item]).to.equal(item);
      }
    }));
  });
});
