// import * as _             from 'lodash';
import { makeInjectable } from './inject';
import { ObjectUtils }    from '../constants/utils/object';

const objectUtils = new ObjectUtils();

const hotCache: any = {};

export interface IComponentOptions extends angular.IDirective {
  selector?: string;
  bind?: any;
  wrapTemplate?: boolean;
}

export const Component = (options: IComponentOptions) => {
  const decorator: ClassDecorator = (Target: any) => {
    const name = options.name || options.selector;

    const dashName = _.kebabCase(name);
    const camelName = _.camelCase(name);
    const className = _.capitalize(camelName);
    const controllerName = `${className}Controller`;

    const viewOptions = _.find(Target.annotations, (item: any) => {
      return item.type === 'view';
    }) || {};

    options = _.clone(options);

    const defaultOptions: IComponentOptions = {
      selector: dashName,
      name: camelName,
      restrict: 'E',
      // TODO: allow provider for customizing this
      controllerAs: 'view',
      scope: {},
      bindToController: options.bind || {},
      controller: controllerName
    };

    _.defaults(options, defaultOptions, viewOptions.value);

    if (!Target.annotations) {
      Target.annotations = [];
    }

    // TODO: ensure component + controller all subclassable
    Target.annotations.push({
      type: 'component',
      decorator: Component,
      value: options
    });

    const injectableOptions = [() => options];

    // TODO: put somewhere else
    const injectableTarget = makeInjectable(Target, false);

    if (hotCache[name]) {
      objectUtils.replace(hotCache[name].injectableTarget, injectableTarget);
      _.extend(hotCache[name].options, options);
    } else {
      // Construct the module
      // TODO: allow link and other functions with imports for directives (?)
      angular
        .module('app')
        .directive(camelName, injectableOptions);

      angular
        .module('app')
        .controller(controllerName, injectableTarget);

      hotCache[name] = { injectableOptions, injectableTarget, options };
    }

    // TODO: do constants get replaced when defined a second time?
    angular
      .module('app')
      .constant(`${className}Component`, Target);

    return Target;
  };

  return decorator;
};
