import { makeInjectable } from './inject';

/**
 * @decorator Controller
 * @description create angular controllers
 * @param {string} name     name of your controller
 * @type {Function}
 * @example
 * ```ts
 * @Controller('TodoController')
 * export class Router {
 *   @inject str: Str;
 *
 * 	myMethod(name: string) {
 * 		str.dasherize(name);
 * 	  // do stuff...
 * 	}
 * }
 * ```
 */
export const Controller = (name: string) => {
  const decorator: ClassDecorator = (Target: any) => {
    const target = makeInjectable(Target);

    // Construct the module
    angular
      .module('app')
      .controller(name, target);

    // Save the controller as a factory too
    angular
      .module('app')
      .constant(name, target);
  };
};
