import { makeInjectable } from './inject';

/**
 * @decorator Directive
 *
 * @description
 * create angular directives from classes
 *
 * @param {string} name
 * name of your directive
 *
 * @example
 * ```ts
 * @Directive('Router')
 * export class Router {
 *   @inject str: Str;
 *
 * 	link(element) {
 * 		str.dasherize(element.text());
 * 	  // do stuff...
 * 	}
 * }
 * ```
 */
export const Directive = (name: string) => {
  const decorator: ClassDecorator = (Target: any) => {
    const target = makeInjectable(Target);

    // Construct the module
    angular
      .module('app')
      .directive(name, target);
  };

  return decorator;
};
