import { makeInjectable } from './inject';

/**
 * @decorator Factory
 *
 * @description create angular services
 *
 * @param {string} name
 * name of your factory
 *
 * @example
 * ```ts
 * @Factory('abTests')
 * export class AbTests {
 *   @inject str: Str;
 *
 * 	 onstructor() {
 * 	   AbTests.str.dasherize('foo');
 * 	 }
 * }
 * ```
 */
export const Factory = (name: string) => {
  const decorator: ClassDecorator = (Target: any) => {
    Target = makeInjectable(Target, true);

    // Construct the module
    angular
      .module('app')
      .factory(name, Target);
  };
};
