/**
 * @decorator inject
 *
 * @type {PropertyDecorator}
 *
 * @todo Allow this to take options like @inject('name') so can alias?
 */
export const inject: PropertyDecorator = (target, name) => {
  const { constructor } = target;

  constructor.$inject = (constructor.$inject || []).slice();
  constructor.$inject.push(name.toString());
};

/**
 * Make a class angular injectable
 *
 * @param {any}     Target
 * @param {boolean} createInstance  Should an instance of this class be created
 *
 * @todo: find a better place for this
 * @todo: make a class or decorator that is subclassed or used on decorator?
 */
export const makeInjectable = (Target: any, createInstance = false, bindContext = false) => {
  const injectList: string[] = Target.$inject || [];

  // Take in services injected by angular and attach
  // them to the class instance
  const init = (...args: any[]) => {
    injectList.forEach((dependencyName, index) => {
      Target.prototype[dependencyName] = args[index];
    });

    if (createInstance) {
      return Target;
    } else {
      const instance = new Target();

      // In some cases class instace methods might be executed
      // without their instance context - like in the case of angular
      // interceptor methods, so we can optionally override
      // context if needed
      if (bindContext) {
        for (const key in instance) {
          const value = instance[key];
          if (typeof value === 'function') {
            instance[key] = value.bind(instance);
          }
        }
      }
      
      return instance;
    }
  };

  // Construct the angular annotation notation for this class
  // e.g. ['myService', (myService) => {}]
  const conf: any[] = injectList.slice();
  conf.push(init);

  return conf;
};
