import { makeInjectable } from './inject';

/**
 * @decorator Config
 *
 * @type {ClassDecorator}
 *
 * @example
 *
 * ```ts
 * @Interceptor
 * export class Routes {
 * 	 @inject
 * 	 private someDependency: SomeDependency;
 *
 * 		// Constructor takes no args for config
 * 		// classes
 * 		request(config: angular.IRequestConfig) {
 * 			if (this.someDependency.foo()) {
 * 				config.params.bar = 'baz';
 * 			}
 * 		}
 * }
 * ```
 */
export const Interceptor: ClassDecorator = (Target: any) => {
  const target = makeInjectable(Target, false, true);

  // Construct the module
  angular
    .module('app')
    .config(['$httpProvider', ($httpProvider: angular.IHttpProvider) => {
      $httpProvider.interceptors.push(target);
    }]);
};
