import { makeInjectable } from './inject';

/**
 * @decorator Service
 *
 * @description create angular services
 *
 * @param {string} name
 * name of your service
 *
 * @example
 * ```ts
 * @Service('abTests')
 * export class AbTests {
 *   @inject str: Str;
 *
 * 	 onstructor() {
 * 	   // do something...
 * 	   str.dasherize('foobar');
 * 	 }
 * }
 * ```
 */
export const Provider = (name: string) => {
  const decorator: ClassDecorator = (Target: any) => {
    const target = makeInjectable(Target, false);

    // Construct the module
    angular
      .module('app')
      .provider(name, target);
  };

  return decorator;
};
