import { Service } from '../decorators/service';
import { inject }  from '../decorators/inject';

@Service('hotReload')

/**
 * @class HotReload
 */
export class HotReload {
  @inject
  $compile: angular.ICompileService;

  /**
   * Find all instances on the current page of a component by
   * name and reload them
   *
   * @param  {string} componentName  Name of the compnoent to reload
   */
  reloadComponent(componentName: string) {
    const selector = _.kebabCase(componentName);

    []
      .forEach
      .call(document.querySelectorAll(selector), (element: Element) => {
        const el = angular.element(element);
        const parentScope: any = el.scope();

        // TODO: may break with ng-if, ng-repeat, etc
        const scope = parentScope.$$childHead;
        const parent = element.parentElement;
        const newElement = el.empty().clone()[0];

        parent.insertBefore(newElement, element);
        el.remove();
        scope.$destroy();

        // TODO: require original directive template instead and compile that?
        this.$compile(newElement)(parentScope);
      });
  }
}
