import { UrlUtils } from '../constants/utils/url';
import { Service } from '../decorators/service';
import { inject }  from '../decorators/inject';
import * as _      from 'lodash';

// Tell the typescript compiler this exists - it is injected
// by webpack
declare var __webpack_require__: any;

@Service('lazy')

/**
 * @class Lazy
 */
export class Lazy {
  @inject
  $ocLazyLoad: oc.ILazyLoad;

  @inject
  $ionicLoading: ionic.loading.IonicLoadingService;

  @inject
  urlUtils: UrlUtils;

  @inject
  $window: any;

  static noop() {
    // Do nothing
  }

  // TODO: also allow route alias config (GUI too?)
  load(name: string) {
    name = _.kebabCase(name);
    const hash: string = _.result(__webpack_require__, 'h');

    // TODO: show ionic spinner after 1 second if not loaded yet
    // until loaded

    // TODO: config for public directory prefix
    // TODO: use path instead of just name because could have components
    // of same name across repos

    let pathPrepend = '/'

    // Workaround for lazy loading while in cordova/mobile mode. Components need the full path.
    if (this.$window.ionic && (this.$window.ionic.Platform.isIOS() || this.$window.ionic.Platform.isAndroid())) {
      pathPrepend = this.urlUtils.getAbsolutePathForCordova();
    }

    const loadedPromise = this.$ocLazyLoad.load(`${pathPrepend}${name}-${hash}.js`);

    this.$ionicLoading.show();
    return loadedPromise.then(() => {
      this.$ionicLoading.hide();
    });
  }
}
