import { Service }   from '../decorators/service';
import { inject }    from '../decorators/inject';
import { Component } from '../services/component';
import * as _        from 'lodash';

@Service('modal')

/**
 * Test 2
 */
export class Modal {
  @inject
  private $ionicModal: ionic.modal.IonicModalService;

  @inject
  private $rootScope: angular.IRootScopeService;

  @inject
  private component: Component;

  show(viewName: string, viewOptions: any = {}, modalOptions: ionic.modal.IonicModalOptions = {}) {
    const camelName = _.camelCase(viewName);
    const template = this.template(camelName, viewOptions);
    const scope: any = modalOptions.scope || this.$rootScope.$new();

    _.defaults(modalOptions, {
      className: _.kebabCase(viewName) + '-modal'
    });

    _.assign(scope, viewOptions);

    const modal = scope.modal = this.$ionicModal.fromTemplate(template, modalOptions);
    modal.show();

    return modal;
  }

  // TODO: move to directive
  template(name: string, options: any = {}) {
    const componentTemplate = this.component.template(name, options);
    const titleName = options && options.title || _.startCase(name);
    const className = options && options.className || '';

    // TODO: allow overridable with config / config provider
    const viewTemplate = `
      <ion-modal-view class="${className}">
        <ion-header-bar>
          <h1 class="title">
            ${titleName}
          </h1>
          <button class="button button-clear" ng-click="modal.hide()">
            Close
          </button>
        </ion-header-bar>
        <ion-content>
          ${componentTemplate}
        </ion-content>
      </ion-modal-view>
    `;

    return viewTemplate;
  }
}
