import { Service }   from '../decorators/service';
import { inject }    from '../decorators/inject';
import { Component } from '../services/component';
import * as _        from 'lodash';

@Service('popover')

/**
 * Popover
 */
export class Popover {
  @inject
  $rootScope: angular.IRootScopeService;

  @inject
  $ionicPopover: ionic.popover.IonicPopoverService;

  @inject
  component: Component;

  show(
    viewName: string,
    event: MouseEvent,
    viewOptions: any = {},
    popoverOptions: ionic.popover.IonicPopoverOptions = {}
  ) {
    const camelName = _.camelCase(viewName);
    const template = this.template(camelName, viewOptions);
    const scope: any = popoverOptions.scope || this.$rootScope.$new();

    const popover = scope.popover = this
      .$ionicPopover
      .fromTemplate(template, popoverOptions);

    popover.show(event);

    return popover;
  }

  template(name: string, options: any = {}) {
    const template = this.component.template(name, options);
    const titleName = options.title || _.startCase(name);

    const popoverTemplate = `
      <ion-popover-view>
        <ion-header-bar>
          <h1 class="title">
            ${titleName}
          </h1>
        </ion-header-bar>
        <ion-content>
          ${template}
        </ion-content>
      </ion-popover-view>
    `;

    return popoverTemplate;
  }
}
