import './dependencies';
import * as http from 'http';
import * as https from 'https';
import * as fs from 'fs';
import * as path  from 'path';

const boot = require('loopback-boot');

export class ShopStyleNodeServer {
  app: any;
  loopback: any;
  config: any;

  private enableSsl = false;
  private sslOptions: https.ServerOptions = {};

  constructor(config: any, loopback: any) {
    this.loopback = loopback;
    this.config = config;
    this.app = this.loopback();
    this.app.config = config;
    this.app.set('config', config);

    this.enableSsl = this.config.sslConfig && this.config.sslConfig.useSsl;
    if (this.enableSsl) {
      this.sslOptions.key = this.readSslFile(this.config.sslConfig.keyFile);
      this.sslOptions.cert = this.readSslFile(this.config.sslConfig.certFile);
    }

    this.boot();
    return this.app;
  }

  private readSslFile(filePath: string): string {
    return fs.readFileSync(path.join(__dirname, '../..', filePath), 'utf-8').toString();
  }

  private boot() {
    // TODO: get from configs
    const bootPath = path.join(process.cwd(), 'server');

    // Bootstrap the application, configure models, datasources and middleware.
    // Sub-apps like REST API are mounted via boot scripts.
    boot(this.app, bootPath, (err: Error) => {
      if (err) {
        throw err;
      }
      this.start();
    });
  }

  private start() {
    let server: http.Server | https.Server;

    if (this.enableSsl) {
      server = https.createServer(this.sslOptions, this.app);
    } else {
      server = http.createServer(this.app);
    }

    // start the web server
    server.listen(this.app.get('port'), () => {
      const baseUrl = (this.enableSsl ? 'https://' : 'http://')
        + this.app.get('host')
        + ':' + this.app.get('port');
      this.app.emit('started', baseUrl);
      console.log('Web server listening at: ' + baseUrl + '...');
    });
  }
}
